import dateutil.parser
from .api import Api


class Vehicle(object):
    def __init__(self, vehicle_id, access_token, unit_system="metric"):
        """Initializes a new Vehicle to use for making requests to the Smartcar API.

        Args:
            vehicle_id (str): the vehicle's unique identifier
            access_token (str): a valid access token
            unit_system (str, optional): the unit system to use for vehicle data.
                Defaults to metric.

        """
        self.vehicle_id = vehicle_id
        self.access_token = access_token
        self.api = Api(access_token, vehicle_id)
        self.api.set_unit_system("metric" if unit_system == "metric" else "imperial")

    def set_unit_system(self, unit_system):
        """Update the unit system to use in requests to the Smartcar API.

        Args:
            unit_system (str): the unit system to use (metric/imperial)

        """
        if unit_system not in ("metric", "imperial"):
            raise ValueError("unit must be either metric or imperial")
        else:
            self.api.set_unit_system(unit_system)

    def info(self):
        """GET Vehicle.info

        Returns:
            dict: vehicle's info

        Raises:
            SmartcarException

        """
        response = self.api.get("")

        return response.json()

    def vin(self):
        """GET Vehicle.vin

        Returns:
            str: vehicle's vin

        Raises:
            SmartcarException

        """
        response = self.api.get("vin")

        return response.json()["vin"]

    def permissions(self):
        """GET Vehicle.permissions

        Returns:
            list: vehicle's permissions

        Raises:
            SmartcarException

        """
        response = self.api.permissions()

        return response.json()["permissions"]

    def has_permissions(self, permissions):
        """Checks if vehicle has specified permission(s).

        Args:
            permissions (str or list of str): Permission(s) to check

        Returns:
            boolean: Whether the vehicle has the specified permission(s)
        """
        vehicle_permissions = self.permissions()
        prefix = "required:"

        if isinstance(permissions, list):
            contained = [
                permission.replace(prefix, "", 1) in vehicle_permissions
                for permission in permissions
            ]

            if False in contained:
                return False
            else:
                return True
        else:
            return permissions.replace(prefix, "", 1) in vehicle_permissions

    def disconnect(self):
        """Disconnect this vehicle from the connected application.

        Note: Calling this method will invalidate your access token and you will
        have to have the user reauthorize the vehicle to your application if you
        wish to make requests to it

        Raises:
            SmartcarException

        """
        self.api.disconnect()

    def odometer(self):
        """GET Vehicle.odometer

        Returns:
            dict: vehicle's odometer

        Raises:
            SmartcarException
        """
        response = self.api.get("odometer")

        return {
            "data": response.json(),
            "unit_system": response.headers["sc-unit-system"],
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def fuel(self):
        """GET Vehicle.fuel

        Returns:
            dict: vehicle's fuel status

        Raises:
            SmartcarException

        """
        response = self.api.get("fuel")

        return {
            "data": response.json(),
            "unit_system": response.headers["sc-unit-system"],
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def oil(self):
        """GET Vehicle.oil

        Returns:
            dict: vehicle's oil status

        Raises:
            SmartcarException

        """
        response = self.api.get("engine/oil")

        return {
            "data": response.json(),
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def tire_pressure(self):
        """GET Vehicle.tire_pressure

        Returns:
            dict: vehicle's tire pressure status

        Raises:
            SmartcarException

        """
        response = self.api.get("tires/pressure")

        return {
            "data": {"tires": response.json()},
            "unit_system": response.headers["sc-unit-system"],
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def battery(self):
        """GET Vehicle.battery

        Returns:
            dict: vehicle's battery status

        Raises:
            SmartcarException

        """
        response = self.api.get("battery")

        return {
            "data": response.json(),
            "unit_system": response.headers["sc-unit-system"],
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def battery_capacity(self):
        """GET Vehicle.battery_capacity

        Returns:

        Raises:
            SmartcarException
        """
        response = self.api.get("battery/capacity")

        return {
            "data": response.json(),
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def charge(self):
        """GET Vehicle.charge

        Returns:
            dict: vehicle's charge status

        Raises:
            SmartcarException

        """
        response = self.api.get("charge")

        return {
            "data": response.json(),
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def location(self):
        """GET Vehicle.location

        Returns:
            dict: vehicle's location

        Raises:
            SmartcarException

        """
        response = self.api.get("location")

        return {
            "data": response.json(),
            "age": dateutil.parser.parse(response.headers["sc-data-age"]),
        }

    def unlock(self):
        """POST Vehicle.unlock

        Returns:
            array:

        Raises:
            SmartcarException

        """
        response = self.api.action("security", "UNLOCK")
        return {"status": response.json()["status"]}

    def lock(self):
        """POST Vehicle.lock

        Returns:
            dict: status

        Raises:
            SmartcarException

        """
        response = self.api.action("security", "LOCK")
        return {"status": response.json()["status"]}

    def start_charge(self):
        """POST Vehicle.start_charge

        Returns:
            array:

        Raises:
            SmartcarException

        """
        response = self.api.action("charge", "START")
        return {"status": response.json()["status"]}

    def stop_charge(self):
        """POST Vehicle.stop_charge

        Returns:
            dict: status

        Raises:
            SmartcarException

        """
        response = self.api.action("charge", "STOP")
        return {"status": response.json()["status"]}

    def batch(self, paths):
        """POST Vehicle.batch

        Args:
            paths (str[]): an array of paths to make
            the batch request to

        Returns:
            dict: the HTTP responses, keyed by path

        Raises:
            SmartcarException

        """
        requests = []
        for path in paths:
            requests.append({"path": path})
        response = self.api.batch(requests)
        batch_dict = dict()
        for response in response.json()["responses"]:
            path = response["path"]
            batch_dict[path] = {
                "code": response["code"],
                "headers": response["headers"],
                "body": response["body"],
            }
        return batch_dict



from . import api, const, requester, vehicle
import time
from datetime import datetime, timedelta
from urllib.parse import urlencode

VERSION = "1.0"


def set_expiration(access):
    expire_date = datetime.utcnow() + timedelta(seconds=access["expires_in"])
    refresh_expire_date = datetime.utcnow() + timedelta(days=60)
    access["expiration"] = expire_date
    access["refresh_expiration"] = refresh_expire_date
    return access


def is_expired(expiration):
    """Check if an expiration is expired

    Args:
        expiration (datetime): expiration datetime

    Returns:
        bool: true if expired
    """
    return datetime.utcnow() > expiration


def set_api_version(version):
    """Update the version of Smartcar API you are using

    Args:
        version (str): the version of the api you want to use
    """
    global VERSION
    VERSION = version


def get_vehicle_ids(access_token, limit=10, offset=0):
    """Get a list of the user's vehicle ids

    Args:
        access_token (str): A valid access token from a previously retrieved
            access object
        limit (integer, optional): The number of vehicle ids to return
        offset (integer, optional): The index to start the vehicle list at

    Returns:
        dict: response containing the list of vehicle ids and paging information

    Raises:
        SmartcarException

    """
    return api.Api(access_token).vehicles(limit=limit, offset=offset).json()


def get_user_id(access_token):
    """Retrieve the userId associated with the access_token

    Args:
        access_token (str): Smartcar access token

    Returns:
        str: userId

    Raises:
        SmartcarException

    """
    return api.Api(access_token).user().json()["id"]


class AuthClient(object):
    def __init__(
        self,
        client_id,
        client_secret,
        redirect_uri,
        scope=None,
        test_mode=None,
        development=None,
    ):
        """A client for accessing the Smartcar API

        Args:
            client_id (str): The application id, provided in the application
                dashboard
            client_secret (str): The application secret, provided in the
                application dashboard
            redirect_uri (str): The URL to redirect to after the user accepts
                or declines the application's permissions. This URL must also be
                present in the Redirect URIs field in the application dashboard
            scope (list, optional): A list of permissions requested by the application
            test_mode (bool, optional): Launch the Smartcar auth flow in test mode. Defaults to false.
            development (bool, optional): DEPRECATED Launch the Smartcar auth flow in development mode
                to enable mock vehicle brands.

        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth = (client_id, client_secret)
        self.redirect_uri = redirect_uri
        self.scope = scope

        if development:
            import warnings

            message = """Development flag is deprecated. This is discouraged and will be
                         removed in the next major release. Use testMode instead."""
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            self.test_mode = development
        else:
            self.test_mode = test_mode if test_mode else False

    def get_auth_url(
        self, force=False, state=None, vehicle_info=None, single_select=None, flags=None
    ):
        """Generate the Connect URL

        Args:
            force (bool, optional): Set to True in order to force the approval
                dialog shown to the user. Defaults to False.
            state (bool, optional): A random string that will be passed back on
                redirect, this allows protection against cross-site forgery
                requests. Defaults to None.
            vehicle_info (dict, optional): A dict with a property, make. Allows
                users to bypass the car brand selection screen, allowing the
                user to go directly to the vehicle login screen.
                Defaults to None.
            single_select (bool or dictionary, optional): An optional value that
                sets the behavior of the grant dialog displayed to the user. It
                can be either a bool or dict. If set to True, `single_select`
                limits the user to selecting only one vehicle. If `single_select`
                is a dictionary with the property `vin`, Smartcar will only authorize the vehicle
                with the specified VIN. See the [Single Select guide](https://smartcar.com/docs/guides/single-select/)
                for more information. Defaults to None.
            flags (str[], optional): List of feature flags that your application has early access to.

        Returns:
            str: authorization url

        Raises:
            SmartcarException

        """
        base_url = const.CONNECT_URL

        approval_prompt = "force" if force else "auto"
        query = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "approval_prompt": approval_prompt,
        }

        if self.test_mode:
            query["mode"] = "test"

        if self.scope:
            query["scope"] = " ".join(self.scope)

        if state:
            query["state"] = state

        if vehicle_info:
            valid_parameters = ["make"]
            for param in valid_parameters:
                if param in vehicle_info:
                    query[param] = vehicle_info[param]

        if single_select is not None:
            query["single_select"] = False
            if isinstance(single_select, dict):
                valid_parameters = ["vin"]
                for param in valid_parameters:
                    if param in single_select:
                        query["single_select_" + param] = single_select[param]
                        query["single_select"] = True
            else:
                query["single_select"] = single_select == True

        if flags:
            query["flags"] = " ".join(flags)

        return base_url + "/oauth/authorize?" + urlencode(query)

    def exchange_code(self, code):
        """Exchange an authentication code for an access dictionary

        Args:
            code (str): A valid authorization code

        Returns:
            dict: dict containing the access and refresh token

        Raises:
            SmartcarException

        """
        method = "POST"
        url = const.AUTH_URL
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.redirect_uri,
        }
        response = requester.call(method, url, data=data, auth=self.auth).json()
        return set_expiration(response)

    def exchange_refresh_token(self, refresh_token):
        """Exchange a refresh token for a new access dictionary

        Args:
            refresh_token (str): A valid refresh token from a previously retrieved
                access object

        Returns:
            dict: dict containing access and refresh token

        Raises:
            SmartcarException

        """
        method = "POST"
        url = const.AUTH_URL
        data = {"grant_type": "refresh_token", "refresh_token": refresh_token}
        response = requester.call(method, url, data=data, auth=self.auth).json()
        return set_expiration(response)

    def is_compatible(self, vin, scope, country="US"):
        """Determine if a vehicle is compatible with Smartcar

        Args:
            vin (str): the VIN of the vehicle
            scope (list): list of permissions to return compatibility info for
            country (str, optional): country code according to [ISO 3166-1 alpha-2](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) of the provided vin

        Returns:
            boolean: true if the vehicle is compatible

        Raises:
            SmartcarException

        """
        method = "GET"
        url = "{}/v{}/compatibility".format(const.API_URL, VERSION)
        query = {"vin": vin, "scope": " ".join(scope), "country": country}

        response = requester.call(method, url, params=query, auth=self.auth).json()
        return response["compatible"]



import platform
import requests
from . import exceptions as E
from . import __version__


def call(method, url, **kwargs):
    """Attachs the kwargs into the headers, sends the request to the Smartcar API
        and handles all error cases

    Args:
        method (str): HTTP method
        url (str): url of the request
        **kwargs: parameters for the request

    Returns:
        dict: response from the request to the Smartcar API

    """
    if "headers" not in kwargs:
        kwargs["headers"] = {}
    kwargs["headers"]["User-Agent"] = "Smartcar/{} ({}; {}) Python v{}".format(
        __version__, platform.system(), platform.machine(), platform.python_version()
    )

    try:
        response = requests.request(method, url, timeout=310, **kwargs)
        code = response.status_code

        if response.ok:
            return response
        elif "/v2.0/" in url:
            raise E.SmartcarExceptionV2(response)
        elif code == 400:
            raise E.ValidationException(response)
        elif code == 401:
            raise E.AuthenticationException(response)
        elif code == 403:
            raise E.PermissionException(response)
        elif code == 404:
            raise E.ResourceNotFoundException(response)
        elif code == 409:
            raise E.StateException(response)
        elif code == 429:
            raise E.RateLimitingException(response)
        elif code == 430:
            raise E.MonthlyLimitExceeded(response)
        elif code == 460:
            raise E.GatewayTimeoutException(response)
        elif code == 500:
            raise E.ServerException(response)
        elif code == 501:
            body = response.json()
            if body["error"] == "smartcar_not_capable_error":
                raise E.SmartcarNotCapableException(response)
            raise E.VehicleNotCapableException(response)
        elif code == 504:
            raise E.GatewayTimeoutException(response)
        else:
            response.raise_for_status()

    except Exception as e:
        if isinstance(e, E.SmartcarException):
            raise e
        else:
            raise E.SmartcarException("Unexpected error") from e



from . import const, requester, smartcar


class Api(object):
    def __init__(self, access_token, vehicle_id=None):
        """Initialize a new Api object to directly make requests to Smartcar.

        Args:
            access_token (str): Smartcar access token
            vehicle_id (str, optional): id of the vehicle. Defaults to None.
        """
        self.access_token = access_token
        self.vehicle_id = vehicle_id
        self.auth = {"Authorization": "Bearer {}".format(access_token)}
        self.unit_system = "metric"

    def set_unit_system(self, unit_system):
        """Update the unit system to use in requests to the Smartcar API.

        Args:
            unit_system (str): the unit system to use (metric/imperial)

        """
        self.unit_system = unit_system

    def _format(self, endpoint):
        """Generates the formated URL

        Args:
            endpoint (str): the Smartcar endpoint of interest

        Returns:
            str: formatted url

        """
        return "{}/v{}/vehicles/{}/{}".format(
            const.API_URL, smartcar.VERSION, self.vehicle_id, endpoint
        )

    def action(self, endpoint, action, **kwargs):
        """Sends POST requests to Smartcar API

        Args:
            endpoint (str): the Smartcar endpoint of interest
            action (str): action to be taken
            **kwargs: information to put into the body of the request

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = self._format(endpoint)
        headers = self.auth
        json = {"action": action}
        for k, v in kwargs.items():
            if v:
                json[k] = v

        return requester.call("POST", url, json=json, headers=headers)

    def batch(self, requests):
        """Sends POST requests to Smartcar API

        Args:
            requests (object[]) - an array of objects containing a 'path' key

        Returns:
            Response: response from the request to the Smartcar API

        """
        endpoint = "batch"
        url = self._format(endpoint)
        json = {"requests": requests}
        headers = self.auth
        headers[const.UNIT_SYSTEM_HEADER] = self.unit_system

        return requester.call("POST", url, json=json, headers=headers)

    def get(self, endpoint):
        """Sends GET requests to Smartcar API

        Args:
            endpoint (str): the Smartcar endpoint of interest

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = self._format(endpoint)
        headers = self.auth
        headers[const.UNIT_SYSTEM_HEADER] = self.unit_system
        return requester.call("GET", url, headers=headers)

    def permissions(self, **params):
        """Sends a request to /permissions

        Args:
            **params: parameters for the request

        Returns:
            Reponse: response from the request to the Smartcar API

        """
        url = self._format("permissions")
        return requester.call("GET", url, headers=self.auth, params=params)

    def disconnect(self):
        """Sends a request to /application

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = self._format("application")
        return requester.call("DELETE", url, headers=self.auth)

    def vehicles(self, **params):
        """Sends a request to /vehicles

        Args:
            **params: parameters for the request

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = "{}/v{}/{}".format(const.API_URL, smartcar.VERSION, "vehicles")
        return requester.call("GET", url, headers=self.auth, params=params)

    def user(self, **params):
        """Sends a request to /user

        Args:
            **params: parameters for the request

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = "{}/v{}/{}".format(const.API_URL, smartcar.VERSION, "user")
        return requester.call("GET", url, headers=self.auth, params=params)



import requests


class SmartcarException(Exception):
    def __init__(self, response):
        self.message = "Unknown error"
        if type(response) is requests.models.Response:
            content_type = response.headers["Content-Type"]
            if "application/json" in content_type:
                json = response.json()
                self.request_id = response.headers.get("sc-request-id")
                if "message" in json:
                    self.message = json["message"]
                elif "error_description" in json:
                    self.message = json["error_description"]
            elif "text/html" in content_type:
                self.message = response.text
        elif type(response) is str:
            self.message = response

    def __str__(self):
        return self.message


class SmartcarExceptionV2(SmartcarException):
    """ Exceptions throw by v2.0 endpoints """

    def __init__(self, response):
        self.type = None
        if type(response) is requests.models.Response:
            content_type = response.headers["Content-Type"]
            if "application/json" in content_type:
                json = response.json()
                self.type = json["type"]
                self.code = json["code"]
                self.description = json["description"]
                self.doc_url = json["docURL"]
                self.status_code = json["statusCode"]
                self.request_id = json["requestId"]
                self.resolution = json["resolution"]
                self.detail = json["detail"]
            elif "text/html" in content_type:
                self.description = response.text
        elif type(response) is str:
            self.description = response

    def __str__(self):
        if self.type is not None:
            return "{}:{} - {}".format(self.type, self.code, self.description)
        return self.description


class ValidationException(SmartcarException):
    pass


class AuthenticationException(SmartcarException):
    pass


class PermissionException(SmartcarException):
    pass


class ResourceNotFoundException(SmartcarException):
    pass


class StateException(SmartcarException):
    def __init__(self, response):
        super(StateException, self).__init__(response)
        json = response.json()
        self.code = json["code"]

    def __str__(self):
        return self.code + ": " + self.message


class RateLimitingException(SmartcarException):
    pass


class MonthlyLimitExceeded(SmartcarException):
    pass


class ServerException(SmartcarException):
    pass


class VehicleNotCapableException(SmartcarException):
    pass


class SmartcarNotCapableException(SmartcarException):
    pass


class GatewayTimeoutException(SmartcarException):
    def __init__(self, response):
        self.message = response.text

    def __str__(self):
        return self.message



from setuptools import setup
import re


def _get_version():
    """Extract version from package."""
    with open("smartcar/__init__.py") as reader:
        match = re.search(
            r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', reader.read(), re.MULTILINE
        )
        if match:
            return match.group(1)
        else:
            raise RuntimeError("Unable to extract version.")


def _get_long_description():
    """Get README contents."""
    with open("README.md") as reader:
        return reader.read()


setup(
    name="smartcar",
    version=_get_version(),
    description="Smartcar Python SDK",
    long_description=_get_long_description(),
    long_description_content_type="text/markdown",
    author="Smartcar",
    author_email="hello@smartcar.com",
    packages=["smartcar"],
    url="https://github.com/smartcar/python-sdk",
    license="MIT",
    install_requires=[
        "python-dateutil",
        "requests",
    ],
    extras_require={
        "dev": [
            "black",
            "coverage",
            "nose",
            "mock",
            "responses",
            "selenium",
            "retrying",
        ]
    },
)



__version__ = "5.2.0"

from .smartcar import (
    AuthClient,
    is_expired,
    get_user_id,
    get_vehicle_ids,
    set_api_version,
)
from .vehicle import Vehicle
from .exceptions import (
    SmartcarException,
    SmartcarExceptionV2,
    ValidationException,
    AuthenticationException,
    PermissionException,
    ResourceNotFoundException,
    StateException,
    RateLimitingException,
    MonthlyLimitExceeded,
    ServerException,
    VehicleNotCapableException,
    SmartcarNotCapableException,
    GatewayTimeoutException,
)
